from dataclasses import asdict, dataclass
from itertools import product
from typing import Any, Optional

import pytest

from ..constants import PREFIX, RESET, SUFFIX, TRUEPREFIX, true_color_supported


@dataclass
class ColorTestCase:
    func_name: str
    message: Any = ''
    color: Optional[bool] = None
    truecolor: Optional[bool] = true_color_supported()
    expected: str = ''

    def __str__(self):
        return ' '.join(f'{key}={value}' for key, value in asdict(self).items())


test_cases = [
    # standard uses
    ColorTestCase('black', message='hi', color=True, truecolor=False, expected=f'{PREFIX}30{SUFFIX}hi{RESET}'),
    ColorTestCase('red', message='hi', color=True, truecolor=False, expected=f'{PREFIX}31{SUFFIX}hi{RESET}'),
    ColorTestCase('green', message='hi', color=True, truecolor=False, expected=f'{PREFIX}32{SUFFIX}hi{RESET}'),
    ColorTestCase('yellow', message='hi', color=True, truecolor=False, expected=f'{PREFIX}33{SUFFIX}hi{RESET}'),
    ColorTestCase('blue', message='hi', color=True, truecolor=False, expected=f'{PREFIX}34{SUFFIX}hi{RESET}'),
    ColorTestCase('magenta', message='hi', color=True, truecolor=False, expected=f'{PREFIX}35{SUFFIX}hi{RESET}'),
    ColorTestCase('cyan', message='hi', color=True, truecolor=False, expected=f'{PREFIX}36{SUFFIX}hi{RESET}'),
    ColorTestCase('white', message='hi', color=True, truecolor=False, expected=f'{PREFIX}37{SUFFIX}hi{RESET}'),

    # truecolor uses
    ColorTestCase('black', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}30;30;30{SUFFIX}hi{RESET}'),
    ColorTestCase('blue', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}0;0;170{SUFFIX}hi{RESET}'),
    ColorTestCase('cyan', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}0;170;170{SUFFIX}hi{RESET}'),
    ColorTestCase('green', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}0;170;0{SUFFIX}hi{RESET}'),
    ColorTestCase('magenta', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}170;0;170{SUFFIX}hi{RESET}'),
    ColorTestCase('red', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}170;0;0{SUFFIX}hi{RESET}'),
    ColorTestCase('yellow', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}170;170;0{SUFFIX}hi{RESET}'),
    ColorTestCase('white', message='hi', color=True, truecolor=True, expected=f'{TRUEPREFIX}170;170;170{SUFFIX}hi{RESET}'),

    # test uncolored uses
    ColorTestCase('black', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('blue', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('cyan', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('green', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('grey', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('magenta', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('red', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('yellow', message='hi', color=False, truecolor=False, expected='hi'),
    ColorTestCase('white', message='hi', color=False, truecolor=False, expected='hi'),
    ]


@pytest.mark.parametrize('test_case', test_cases, ids=list(map(str, test_cases)))
def test_colors(test_case):
    import termlog

    func = getattr(termlog, test_case.func_name)
    result = func(message=test_case.message, color=test_case.color, truecolor=test_case.truecolor)
    assert result == test_case.expected, f'TrueColor={test_case.truecolor} ({tuple(result)}) != ({tuple(test_case.expected)})'


@dataclass
class RgbColorTestCase:
    red: int = 0
    green: int = 0
    blue: int = 0
    message: Any = ''
    color: Optional[bool] = None
    truecolor: Optional[bool] = None

    def __str__(self):
        return ' '.join(f'{key}={value}' for key, value in asdict(self).items())


reds = greens = blues = [0, 127, 255]
truecolors = [True, False]
rgb_test_cases = []
for values in product(reds, greens, blues, truecolors):
    rgb_test_cases.append(RgbColorTestCase(red=values[0], green=values[1], blue=values[2], message='hi', truecolor=values[3]))


@pytest.mark.parametrize('test_case', rgb_test_cases, ids=list(map(str, rgb_test_cases)))
def test_colors_rgb(test_case):
    from .. import colors

    prefix = f'\x1b[38;2;{test_case.red};{test_case.green};{test_case.blue}m'
    reset = f'\x1b[0m'
    expected = f'{prefix}{test_case.message}{reset}' if test_case.truecolor else f'{test_case.message}'

    result = colors.rgb(red=test_case.red, blue=test_case.blue, green=test_case.green, message=test_case.message, color=test_case.color, truecolor=test_case.truecolor)
    assert result == expected



import functools
from dataclasses import dataclass, field
from typing import Any, Dict, List


@dataclass
class factory:
    """Caches classes

    Attributes:
        names: names of class fields to use for caching
        repository: cache dictionary
        capture_subclasses: also cache subclasses

    """
    names: List[str] = field(default_factory=list)
    repository: Dict[Any, Any] = field(repr=False, default_factory=dict)
    capture_subclasses: bool = False

    def __repr__(self):
        return f'{self.__class__.__name__}(names={self.names}, count={len(self.repository)})'

    def generate_key(self, cls, *args, **kwds):
        cls_key = {}
        for index, name in enumerate(self.names):
            if name in kwds:
                value = kwds.get(name)
            elif len(args) > index:
                value = args[index]
            elif hasattr(cls, name):

                value = getattr(cls, name)
            else:
                # Guard for programming errors during development
                raise KeyError(f'Could not find `{name}` in {cls} `__new__` invocation: args={args} kwds={kwds}')
            cls_key[name] = value
        vals = tuple(cls_key.values())
        return vals

    def get_instance(self, cls, *args, **kwds):
        """Looks up instance if possible, creates a new one if the key
        doesn't exist.

        """
        vals = self.generate_key(cls, *args, **kwds)
        if vals not in self.repository:
            instance = object.__new__(cls)
            self.repository[vals] = instance
        instance = self.repository[vals]
        return instance

    def __call__(self, cls, *args, **kwds):
        # print(f'self={self} cls={cls}, args={args}, kwds={kwds}')

        @functools.wraps(cls.__new__)
        def new(cls, *args, **kwds):
            return self.get_instance(cls, *args, **kwds)

        # this works with one decorator but not multiple
        #  TODO: make this wrap factories and other classes
        #        so we can stack factories based on different
        #        attributes
        cls.__new__ = new
        return cls



from typing import Any

from . import palettes
from .__metadata__ import package_metadata
from .colors import Color, rgb
from .formatting import beautify, format
from .palettes import Palette, get_palette, set_palette
from .terminal import echo, set_config

__author__ = package_metadata.author
__version__ = package_metadata.version


def __getattr__(name: str) -> Any:
    """__getattr__ should be called _after_ __getattribute__ and only if
    __getattribute__ could not find the correct value...

    This is magic and will grab the color of the default palette as if
    it were part of the top level termlog structure.

    """
    palette = locals()['palette'] = palettes.get_palette()
    # This handles from termlog import *
    #   But from termlog import * prevents updates to the palette,
    #   And this is by design since imports are technically cached within
    #   pythons import system.  It's all code, so we could force it to
    #   "just work" but that seems like a bad idea.
    #
    # The best practice is to use termlog as a namespace/prefix:
    #
    #    >>> import termlog
    #    >>> import termlog.palettes
    #    >>> termlog.red('hi')
    #    '\x1b[31mhi\x1b[0m'
    #    >>> termlog.red('hi', truecolor=True)
    #    '\x1b[38;2;170;0;0mhi\x1b[0m'
    #    >>> termlog.set_palette(termlog.palettes.SolarizedDark())
    #    >>> termlog.red('hi')
    #    '\x1b[31mhi\x1b[0m'
    #    >>> termlog.red('hi', truecolor=True)
    #    '\x1b[38;2;220;50;47mhi\x1b[0m'
    #
    if name == '__all__':
        return [key for key in palette.__annotations__ if isinstance(getattr(palette, key), Color)] + [
            'Color', 'rgb', 'format', 'beautify', 'get_palette', 'set_palette', 'Palette', 'echo', 'set_config'
            ]
    else:
        return getattr(palette, name)



def test_multilayer_calls():
    """This is harder to setup as a test case, so we only have one"""
    import json
    from termlog import format, red

    message = 'green'
    output = json.loads(format(f'A {red(message)} message!', json=True, color=False, add_timestamp=False))
    assert output == {'data': 'A green message!', 'message': 'green'}



import os


# This will get updated below, see: _true_color_supported
def true_color_supported() -> bool:
    """Check if truecolor is supported by the current tty.

    Note: this currently only checks to see if COLORTERM contains
          one of the following enumerated case-sensitive values:
             - truecolor
             - 24bit

    """
    color_term = os.getenv('COLORTERM', '')
    return True if any(check in color_term for check in ['truecolor', '24bit']) else False


TRUE_COLOR_SUPPORTED = true_color_supported()

ESCAPE: str = '\x1b'
PREFIX: str = f'{ESCAPE}['
TRUEPREFIX: str = f'{PREFIX}38;2;'
SUFFIX: str = 'm'
RESET: str = f'{PREFIX}0{SUFFIX}'

# style codes
#  Not all of these are supported within all terminals
BRIGHT: str = f'{PREFIX}1{SUFFIX}'
DIM: str = f'{PREFIX}2{SUFFIX}'

BLINKING: str = f'{PREFIX}5{SUFFIX}'
STROBING: str = f'{PREFIX}6{SUFFIX}'

UNDERLINED: str = f'{PREFIX}4{SUFFIX}'
DOUBLE_UNDERLINE: str = f'{PREFIX}21{SUFFIX}'

ITALICS: str = f'{PREFIX}3{SUFFIX}'
INVERTED: str = f'{PREFIX}7{SUFFIX}'
HIDDEN: str = f'{PREFIX}8{SUFFIX}'
STRIKE_THROUGH: str = f'{PREFIX}9{SUFFIX}'



"""Color module

Contains resources to build 256 (standard) and 16m (true-color) text
output.

"""
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Tuple

from .constants import (
    BLINKING,
    BRIGHT,
    DIM,
    DOUBLE_UNDERLINE,
    HIDDEN,
    INVERTED,
    ITALICS,
    PREFIX,
    RESET,
    STRIKE_THROUGH,
    STROBING,
    SUFFIX,
    TRUE_COLOR_SUPPORTED,
    TRUEPREFIX,
    UNDERLINED
)
from .decorations import factory

__all__ = ('Color', 'rgb')


_colors: Dict[Tuple[int, int, int], 'Color'] = {}


@factory(names=['red', 'green', 'blue'], repository=_colors)  # look up color by rgb value
@dataclass
class Color:
    """A data structure for packaging up Color display information

    Example:
        >>> from termlog import Color, echo
        >>> solarized_red = Color(red=220, green=50, blue=47)
        >>> solarized_magenta = Color(red=221, green=54, blue=130)
        >>> msg = ...
        >>> echo(f'{solarized_red("ERROR")}: {solarized_magenta(msg)}')

    Attributes:
        red: the value for the red aspect of the color [0-255]
        green: the value for the green aspect of the color [0-255]
        blue: the value for the blue aspect of the color [0-255]
        color_prefix: the string prefix when true-color is not enabled
        true_color_prefix: the string prefix for when true-color is enabled
        suffix: the reset suffix
        truecolor: forces output to use truecolor
        prefix: calculated prefix based on flags

    Flags:
        bright: adds bold/bright flag
        dim: adds dim flag
        italics: adds italics flag
        underlined: adds underlined flag
        double_underlined: adds double underlined flag
        strobing: adds strobing flag
        blinking: adds blinking flag
        inverted: flag that inverts foreground and background colors
        strike_through: adds strike-through flag
        hidden: adds hidden flag

    """
    red: int = 0
    green: int = 0
    blue: int = 0
    term_color: Optional[int] = None
    color_prefix: str = ''
    true_color_prefix: str = f'{TRUEPREFIX}{{red}};{{green}};{{blue}}{SUFFIX}'
    suffix: str = f'{RESET}'
    truecolor: bool = TRUE_COLOR_SUPPORTED
    name: str = ''

    # style attributes
    dim: bool = False
    bright: bool = False
    italics: bool = False
    underlined: bool = False
    double_underlined: bool = False
    blinking: bool = False
    strobing: bool = False
    inverted: bool = False
    hidden: bool = False
    strike_through: bool = False

    # style attributes combined
    postfix: str = field(default='', init=False)

    def __set_name__(self, owner: Any, name: str):
        setattr(owner, name, self)
        self.name = name

    def __post_init__(self):
        # Cap red, green and blue to integer values
        self.red = max(min(int(self.red or 0), 255), 0)
        self.green = max(min(int(self.green or 0), 255), 0)
        self.blue = max(min(int(self.blue or 0), 255), 0)

        if self.term_color is not None and not self.color_prefix:
            self.color_prefix = f'{PREFIX}{self.term_color}{SUFFIX}'

        if self.truecolor is None:
            self.truecolor = TRUE_COLOR_SUPPORTED

        self.true_color_prefix = self.true_color_prefix.format(red=self.red, green=self.green, blue=self.blue)
        if self.true_color_prefix and not self.color_prefix:
            self.color_prefix = self.true_color_prefix
        elif self.color_prefix and not self.true_color_prefix:
            self.true_color_prefix = self.color_prefix

        self.postfix = self.generate_postfix(
            dim=self.dim,
            bright=self.bright,
            italics=self.italics,
            underlined=self.underlined,
            double_underlined=self.double_underlined,
            blinking=self.blinking,
            strobing=self.strobing,
            strike_through=self.strike_through,
            hidden=self.hidden,
            inverted=self.inverted,
            )

    @staticmethod
    def generate_postfix(
            dim: Optional[bool] = None,
            bright: Optional[bool] = None,
            italics: Optional[bool] = None,
            underlined: Optional[bool] = None,
            double_underlined: Optional[bool] = None,
            hidden: Optional[bool] = None,
            inverted: Optional[bool] = None,
            strike_through: Optional[bool] = None,
            blinking: Optional[bool] = None,
            strobing: Optional[bool] = None,
            ) -> str:
        """Creates a prefix given the constraints above

        Notes:
            The following options are grouped and only one flag will be
            honored for each grouping:

            * color, truecolor
            * strobing, blinking
            * dim, bright
            * underlined, double_underlined

            Additionally, if the hidden flag is set, none of the other
            flags will be honored

        Args:
            dim: sets the dim flag
            bright: sets the bright flag
            italics: sets the italics flag
            hidden: hide the output
            inverted: invert the foreground and background colors
            strike_through: set the strike_through flag
            underlined: set the underline flag
            double_underlined: set the underline flag
            blinking: set the blinking flag
            strobing: set the strobing flag

        Returns:
            message with escape sequences for terminal colors

        """
        postfix = ''
        # ignore all of the other flags if hidden
        if hidden:
            postfix += HIDDEN
        else:
            if inverted:
                postfix += INVERTED
            if strike_through:
                postfix += STRIKE_THROUGH
            if italics:
                postfix += ITALICS
            # dim or bright but not both and only if we don't use
            #   truecolor
            if dim:
                postfix += DIM
            elif bright:
                postfix += BRIGHT
            # underline or double underline but not both
            if underlined:
                postfix += UNDERLINED
            elif double_underlined:
                postfix += DOUBLE_UNDERLINE
            # blinking or strobing but not both
            if blinking:
                postfix += BLINKING
            elif strobing:
                postfix += STROBING
        return postfix

    def __hash__(self):
        return hash((self.red, self.green, self.blue))

    def __call__(self, message: Any, color: bool = True, truecolor: Optional[bool] = None) -> str:
        """Colors a message

        Args:
            color: flag to disable color entirely
            truecolor: flag to explicitly enable truecolor

        Returns:
            message with escape sequences for terminal colors

        """
        colored_message = message
        prefix = ''
        if color is not False and (truecolor is True or (truecolor is None and self.truecolor)):
            prefix = self.true_color_prefix + self.postfix
        elif color:
            prefix = self.color_prefix + self.postfix
        if prefix:
            colored_message = f'{prefix}{message}{self.suffix}'
        return colored_message

    def __eq__(self, other):
        return id(self) == id(other)

    def __getitem__(self, item):
        if item in ('red', 'green', 'blue'):
            return getattr(self, item)
        else:
            raise KeyError(f'Could not find {item} in {self}')

    @staticmethod
    def keys():
        return ('red', 'green', 'blue')

    @staticmethod
    def __len__():
        return 3

    def __iter__(self):
        for name in ('red', 'green', 'blue'):
            yield getattr(self, name)


def rgb(message: Any, red: int = 0, green: int = 0, blue: int = 0, color: bool = True, truecolor: Optional[bool] = None) -> str:
    """Add any true color to *message*.

    Example:
        >>> from termlog import echo, rgb, Color
        >>> c = Color(red=223, green=81, blue=127)
        >>> print(f'{c!r}')
        Color(red=223, green=81, blue=127, color_prefix='', true_color_prefix='{TRUEPREFIX}{red};{green};{blue}{SUFFIX}', suffix='{RESET}', truecolor=False)
        >>> f'A {rgb("colored", **c)} message'
        'A {TRUEPREFIX}223;81;127{SUFFIX}colored{RESET} message'

    Note: this only works if the terminal supports true color and there
          is no way to definitively determine if truecolor is supported
          by the terminal.

    Args:
        message: text to color
        red: red portion of color [0 to 255]
        green: green portion of color [0 to 255]
        blue: blue portion of color [0 to 255]
        color: enable color on output

    Returns:
        Colored text if color is enabled

    """
    _color = Color(red=red, green=green, blue=blue, truecolor=True if truecolor is not False else False)
    colored_text = _color(message, color=color)
    return colored_text



import ast
import datetime
import inspect
import json
import re
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Optional


def fix_json(obj):
    try:
        return json._default_decoder(obj)
    except Exception:
        return str(obj)


@dataclass
class Message:
    data: Any
    timestamp: datetime.datetime = field(default_factory=datetime.datetime.utcnow)
    json: bool = False
    color: bool = False
    type: Any = str
    lexer: str = ''
    time_format: Optional[str] = '%Y%m%d%H%M%S'
    include_timestamp: bool = True
    fields: Dict = field(default_factory=dict)

    def __post_init__(self):
        if self.color is False:
            self.data = strip_escape(f'{self.data}')
        self.time_format = '%Y%m%d%H%M%S' if self.time_format is None else self.time_format
        self.timestamp = self.timestamp.strftime(self.time_format)
        self._update_fields()

    @property
    def calling_frame(self):
        found = False
        frame = None
        stopping_functions = ['echo', 'format', 'red']
        for frame in inspect.getouterframes(inspect.currentframe()):
            if found:
                break
            if 'termlog' in frame.filename and frame.function in stopping_functions:
                found = True
        return None if not (frame and found) else frame

    @property
    def calling_frame_code(self):
        frame = self.calling_frame
        code = ''
        if frame and frame.code_context:
            code = ''.join(frame.code_context).strip()
        return code

    @property
    def calling_frame_data(self):
        frame_data = {}
        frame = self.calling_frame
        if frame:
            members = {k: v for k, v in inspect.getmembers(frame)}
            key_frame = members['frame']
            frame_data.update(key_frame.f_globals)
            frame_data.update(key_frame.f_locals)
        return frame_data

    def _update_fields(self):
        frame_data = self.calling_frame_data
        code = self.calling_frame_code
        parsed = ast.parse(code)
        queue = parsed.body
        data = []
        fields = []
        # Grab field names to get data needed for message
        count = -1
        while queue:
            count += 1
            node = queue.pop(0)
            if isinstance(node, (ast.Expr, ast.FormattedValue, ast.Assign, ast.Starred, ast.Attribute)):
                queue.append(node.value)
            elif isinstance(node, (ast.Call, )):
                # TODO: Find a way to capture the colors here
                for arg in node.args:
                    queue.append(arg)
            elif isinstance(node, (ast.JoinedStr, )):
                for value in node.values:
                    queue.append(value)
            elif isinstance(node, (ast.Str, )):
                data.append(node.s)
            elif isinstance(node, (ast.Name, )):
                fields.append(node.id)
            elif isinstance(node, (ast.BinOp, )):
                queue.append(node.left)
                queue.append(node.right)
            else:
                print(node, ', '.join([d for d in dir(node) if not d.startswith('_')]))

            if count > 1000:  # to prevent a runaway
                break
        for name in fields:
            if name in frame_data:
                self.fields[name] = frame_data[name]

    def __eq__(self, other):
        try:
            equal = (self.data == other.data)
        except AttributeError:
            equal = (self.data == other)
        return equal

    def __radd__(self, other) -> 'Message':
        message = self
        try:
            new_fields = {k: v for k, v in self.fields.items()}
            if isinstance(other, Message):
                new_fields.update(other.fields)
                text = other.data + self.data
            else:
                text = other + str(self.data)
            message = Message(
                text,
                timestamp=self.timestamp,
                json=self.json,
                color=self.color,
                lexer=self.lexer,
                time_format=self.time_format,
                include_timestamp=self.include_timestamp,
                fields=new_fields,
                )
        except (AttributeError, TypeError):
            text = other + str(self.data)
            message = Message(
                text,
                timestamp=self.timestamp,
                json=self.json,
                color=self.color,
                lexer=self.lexer,
                time_format=self.time_format,
                include_timestamp=self.include_timestamp,
                fields=self.fields,
                )
        finally:
            return message

    def __add__(self, other) -> 'Message':
        try:
            new_fields = {k: v for k, v in self.fields.items()}
            new_fields.update(other.fields)
            return Message(
                self.data + other.data,
                timestamp=self.timestamp,
                json=self.json,
                color=self.color,
                lexer=self.lexer,
                time_format=self.time_format,
                include_timestamp=self.include_timestamp,
                fields=new_fields,
                )
        except AttributeError:
            return Message(
                self.data + other,
                timestamp=self.timestamp,
                json=self.json,
                color=self.color,
                lexer=self.lexer,
                time_format=self.time_format,
                include_timestamp=self.include_timestamp,
                fields=self.fields,
                )

    def __str__(self):
        from .formatting import beautify
        data = asdict(self)
        # ts = '' if not self.include_timestamp else f'{self.timestamp.strftime(self.time_format)} '
        ts = '' if not self.include_timestamp else f'{self.timestamp} '
        if self.json:
            msg_fields = ['json', 'color', 'lexer', 'type', 'time_format', 'include_timestamp']
            for field_name in msg_fields:
                data.pop(field_name)
            for field_name, field_value in data.pop('fields', {}).items():
                data[field_name] = field_value
            if not self.include_timestamp:
                data.pop('timestamp')
            string = json.dumps(data, default=fix_json)
        elif self.lexer and self.color:
            string = beautify(self.data, lexer=self.lexer)
        else:
            string = f'{self.data}'
        if self.include_timestamp and not self.json:
            string = f'{ts}{string}'
        if not self.color:
            string = strip_escape(string)
        return str(string)


def strip_escape(text: str) -> str:
    """Remove terminal ascii escape sequences from *text*.

    Args:
        text: text to strip escape sequences from

    Returns:
        text stripped of escape sequences

    """
    # These are all valid escape sequences... they're probably not
    #  inclusive
    sequences = [
        r'\033',
        r'\x1b',
        r'\u001b',
        ]
    value_sequence = r'\[[0-9;]+m'
    for escape_sequence in sequences:
        pattern = f'{escape_sequence}{value_sequence}'
        replacement = ''
        text = re.sub(pattern, replacement, text)
    return text



"""Palettes allow someone to create a specific color scheme for the
terminal log.
"""
from dataclasses import asdict, dataclass, field
from typing import Dict

from .colors import Color
from .decorations import factory
from .message import strip_escape

__all__ = ('get_palette', 'set_palette', 'Palette')
_palettes: Dict[str, 'Palette'] = {}


@factory(names=['name'], repository=_palettes)
@dataclass
class Palette:
    """A data structure to capture colors.

    Palettes can be defined which will modify termlog's color
    and output.

    Example:

        >>> import termlog
        >>> termlog.red('hi', truecolor=True)
        '\x1b[38;2;170;0;0mhi\x1b[0m'

        >>> from dataclasses import dataclass, field
        >>> @dataclass
            class NewPalette(termlog.Palette):
                name: str = 'My special palette'

                red: termlog.Color = field(default=termlog.Color(185, 10, 10, term_color=31))

        >>> p = NewPalette()
        >>> termlog.set_palette(p)
        >>> termlog.red('hi', truecolor=True)
        '\x1b[38;2;185;10;10mhi\x1b[0m'

    """
    name: str = ''
    colors: Dict = field(default_factory=dict)

    @staticmethod
    def strip(string) -> str:
        """Strips out escape characters from string"""
        return strip_escape(string)

    def __init_subclass__(cls, **kwargs) -> None:
        if cls.name not in _palettes:
            _palettes[cls.name] = cls()
            super().__init_subclass__(**kwargs)  # type: ignore

    def __post_init__(self):
        for key, value in asdict(self).items():
            if isinstance(value, Color):
                self.colors[key] = value

    def __getattr__(self, key):
        if key in self.colors:
            return self.colors[key]
        else:
            raise AttributeError(f'Could not find `{key}` in `{self}`')


@dataclass
class Default(Palette):
    name: str = 'default'

    black: Color = field(default=Color(30, 30, 30, term_color=30))
    red: Color = field(default=Color(170, 0, 0, term_color=31))
    green: Color = field(default=Color(0, 170, 0, term_color=32))
    yellow: Color = field(default=Color(170, 170, 0, term_color=33))
    blue: Color = field(default=Color(0, 0, 170, term_color=34))
    magenta: Color = field(default=Color(170, 0, 170, term_color=35))
    cyan: Color = field(default=Color(0, 170, 170, term_color=36))
    white: Color = field(default=Color(170, 170, 170, term_color=37))

    bright_black: Color = field(default=Color(60, 60, 60, term_color=90))
    bright_red: Color = field(default=Color(255, 0, 0, term_color=91))
    bright_green: Color = field(default=Color(0, 255, 0, term_color=92))
    bright_yellow: Color = field(default=Color(255, 255, 0, term_color=93))
    bright_blue: Color = field(default=Color(0, 0, 255, term_color=94))
    bright_magenta: Color = field(default=Color(255, 0, 255, term_color=95))
    bright_cyan: Color = field(default=Color(0, 255, 255, term_color=96))
    bright_white: Color = field(default=Color(255, 255, 255, term_color=97))

    dim_black: Color = field(default=Color(0, 0, 0, term_color=30, dim=True))
    dim_red: Color = field(default=Color(85, 0, 0, term_color=31, dim=True))
    dim_green: Color = field(default=Color(0, 85, 0, term_color=32, dim=True))
    dim_yellow: Color = field(default=Color(85, 85, 0, term_color=33, dim=True))
    dim_blue: Color = field(default=Color(0, 0, 85, term_color=34, dim=True))
    dim_magenta: Color = field(default=Color(85, 0, 85, term_color=35, dim=True))
    dim_cyan: Color = field(default=Color(0, 85, 85, term_color=36, dim=True))
    dim_white: Color = field(default=Color(85, 85, 85, term_color=37, dim=True))
    grey: Color = field(default=dim_white)


@dataclass
class SolarizedDark(Palette):
    name: str = 'solarized_dark'

    # Standard terminal names
    black: Color = field(default=Color(7, 54, 66, term_color=30))
    red: Color = field(default=Color(220, 50, 47, term_color=31))
    green: Color = field(default=Color(133, 153, 0, term_color=32))
    yellow: Color = field(default=Color(181, 137, 0, term_color=33))
    blue: Color = field(default=Color(38, 139, 210, term_color=34))
    magenta: Color = field(default=Color(211, 54, 130, term_color=35))
    cyan: Color = field(default=Color(42, 161, 152, term_color=36))
    white: Color = field(default=Color(238, 232, 213, term_color=37))

    # Standard bright terminal names
    bright_black: Color = field(default=Color(0, 43, 54, term_color=90))
    bright_red: Color = field(default=Color(203, 75, 22, term_color=91))
    bright_green: Color = field(default=Color(88, 110, 117, term_color=92))
    bright_yellow: Color = field(default=Color(101, 123, 131, term_color=93))
    bright_blue: Color = field(default=Color(131, 148, 150, term_color=94))
    bright_magenta: Color = field(default=Color(108, 113, 196, term_color=95))
    bright_cyan: Color = field(default=Color(147, 161, 161, term_color=96))
    bright_white: Color = field(default=Color(253, 246, 227, term_color=97))

    # solarized specific names
    base03: Color = field(default=bright_black)
    base02: Color = field(default=black)
    base01: Color = field(default=bright_green)
    base00: Color = field(default=bright_yellow)
    base0: Color = field(default=bright_blue)
    base1: Color = field(default=bright_cyan)
    base2: Color = field(default=white)
    base3: Color = field(default=bright_white)
    orange: Color = field(default=bright_red)
    violet: Color = field(default=bright_magenta)


_default_palette: Palette = Default()


def get_palette() -> Palette:
    return _default_palette


def set_palette(palette: Palette):
    global _default_palette
    _default_palette = palette



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Union

import pytest
from pygments.lexer import Lexer


@dataclass
class EchoTestCase:
    messages: List[Any]
    verbose: Optional[Union[bool, int]] = None
    end: str = '\n'
    flush: bool = True
    lexer: Optional[Union[Lexer, str]] = None
    color: Optional[bool] = None
    json: Optional[bool] = None
    add_timestamp: bool = False
    expected: Optional[str] = None

    @property
    def echo_kwds(self) -> Dict[str, Any]:
        data = asdict(self)
        data.pop('expected')
        data.pop('messages')
        return data


@pytest.mark.parametrize('test_case', [
    EchoTestCase([''], json=False, color=True),
    EchoTestCase(['a']),
    EchoTestCase([1], expected='1'),
    EchoTestCase([None], expected='None'),
    ])
def test_echo(test_case, capfd):
    from .. import echo

    messages = test_case.messages
    expected = test_case.expected if test_case.expected is not None else ' '.join(f'{m}' for m in messages)

    result = echo(*messages, **test_case.echo_kwds)
    cap = capfd.readouterr()
    assert result == expected
    assert cap.err == ''
    # This fails...why?
    # assert cap.out == expected


# TODO:
# @dataclass
# class ConfigTestCase:
#     message: str = ''
#     expected: str = ''
#     color: Optional[bool] = None
#     json: Optional[bool] = None
#     time_format: Optional[str] = None
#     string_format: Optional[str] = None
#
#     @property
#     def echo_kwds(self) -> Dict[str, Any]:
#         data = asdict(self)
#         data.pop('expected')
#         data.pop('message')
#         return data
#
#
# @pytest.mark.parametrize('test_case', [
#     ConfigTestCase('hi', ''),
#     ])
# def test_set_config(test_case, capfd):
#     from .. import echo, set_config
#
#     set_config(**test_case.echo_kwds)
#
#     message = test_case.message
#     expected = test_case.expected if test_case.expected is not None else message
#
#     result = echo(message, add_timestamp=False, **test_case.echo_kwds)
#     cap = capfd.readouterr()
#     assert cap.err == ''
#     assert cap.out == expected
#     assert result == expected



from dataclasses import dataclass
from typing import Optional


@dataclass
class TerminalConfig:
    """Data class to hold some simple configuration data

    Attributes:
        color: when True, display output with color
        json: when True, display output as json
        timestamp: when True, include timestamp in output
        time_format: control how the timestamp appears in the output

    """
    color: bool = True
    json: bool = False
    timestamp: Optional[bool] = None
    time_format: str = '%Y%m%d%H%M%S'


_terminal_config = TerminalConfig()


def set_config(
        color: Optional[bool] = None,
        json: Optional[bool] = None,
        time_format: Optional[str] = None,
        timestamp: Optional[bool] = None,
        ) -> TerminalConfig:
    """Sets configuration for subsequent termlog API calls.

    Note:  Generally, True should only be set for color or json

    Args:
        color: True will enable color
        json: True will enable json output
        time_format: Sets the timestamp format
        timestamp: True will enable timestamps on each log line

    Returns:
        Updated configuration

    """
    global _terminal_config
    _terminal_config.color = color if color is not None else _terminal_config.color
    _terminal_config.json = json if json is not None else _terminal_config.json
    _terminal_config.timestamp = timestamp if timestamp is not None else _terminal_config.timestamp
    _terminal_config.time_format = time_format if time_format is not None else _terminal_config.time_format
    return _terminal_config
